from __future__ import annotations

import json
from datetime import datetime

from pydantic import BaseModel


class Job(BaseModel):
    args: list
    kwargs: dict
    create_at: int = 0
    """
    Timestamp in milliseconds
    """

    uid: str = "no_uid"

    @classmethod
    def from_redis(cls, serialized: str) -> Job:
        return cls.model_validate_json(serialized)

    @classmethod
    def from_message(cls, message: dict) -> Job:
        return cls.model_validate_json(message["payload"])

    def to_redis(self) -> str:
        return json.dumps(self.model_dump(), sort_keys=True)

    def to_message(self) -> dict:
        return {"payload": self.to_redis()}



import os

os.environ["BRQ_LOG_LEVEL"] = "DEBUG"

import socket
import time

import pytest
import redis
import redis.asyncio

import docker
from brq.tools import get_redis_url


def get_port():
    # Get an unoccupied port
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        return s.getsockname()[1]


@pytest.fixture(scope="session")
def docker_client():
    try:
        client = docker.from_env()
        client.ping()
        return client
    except:
        pytest.skip("Docker is not available")


@pytest.fixture(scope="session")
def redis_port(docker_client):
    """
    Start a redis container and return the port
    """
    redis_port = get_port()
    container = None
    try:
        container = docker_client.containers.run(
            f"redis:{os.getenv('REDIS_VERSION', '7')}",
            detach=True,
            ports={"6379": redis_port},
            remove=True,
        )
        time.sleep(1)  # Wait for the server to start
        while True:
            try:
                # Ping redis
                redis_client = redis.Redis(host="localhost", port=redis_port)
                redis_client.ping()
            except:
                time.sleep(0.5)
            else:
                break
        yield redis_port
    finally:
        if container:
            container.stop()


@pytest.fixture
async def async_redis_client(redis_port):
    redis_url = get_redis_url(port=redis_port)
    redis_client = redis.asyncio.from_url(redis_url, decode_responses=True)
    try:
        await redis_client.flushall()
        yield redis_client
    finally:
        await redis_client.aclose()


@pytest.fixture
def stream_name():
    return "test_stream"


@pytest.fixture
def group_name():
    return "test_group"



import asyncio
import contextlib
from contextlib import asynccontextmanager
from datetime import datetime
from typing import Any, AsyncGenerator

import redis.asyncio as redis

from brq.log import logger


async def get_current_timestamp(
    redis_client: redis.Redis | redis.RedisCluster,
) -> int:
    return int((await redis_client.time())[0])


async def get_current_timestamp_ms(
    redis_client: redis.Redis | redis.RedisCluster,
) -> int:
    return int((await redis_client.time())[0] * 1000 + (await redis_client.time())[1] / 1000)


def timestamp_to_datetime(timestamp: int, tz=None) -> datetime:
    return datetime.fromtimestamp(timestamp, tz=tz)


def timestamp_ms_to_datetime(timestamp_ms: int) -> datetime:
    return datetime.fromtimestamp(timestamp_ms / 1000)


async def event_wait(evt, timeout):
    # suppress TimeoutError because we'll return False in case of timeout
    with contextlib.suppress(asyncio.TimeoutError):
        await asyncio.wait_for(evt.wait(), timeout)
    return evt.is_set()


def get_redis_url(
    host: str = "localhost",
    port: int = 6379,
    db: int = 0,
    cluster: bool = False,
    tls: bool = False,
    username: str = "",
    password: str = "",
):
    if not cluster:
        url = f"{host}:{port}/{db}"
    else:
        logger.debug("Cluster mode, no need to specify db")
        url = f"{host}:{port}"

    if username or password:
        url = f"{username}:{password}@{url}"

    if tls:
        return f"rediss://{url}"
    else:
        return f"redis://{url}"


@asynccontextmanager
async def get_redis_client(
    redis_url,
    is_cluster: bool = False,
) -> AsyncGenerator[Any, redis.Redis | redis.RedisCluster]:
    if is_cluster:
        redis_client = redis.RedisCluster.from_url(redis_url, decode_responses=True)
    else:
        redis_client = redis.from_url(redis_url, decode_responses=True)
    try:
        await redis_client.ping()
        logger.debug("Connected to redis")
        yield redis_client
    finally:
        await redis_client.aclose()



import asyncio
import copy
import signal

from brq.consumer import Consumer, RunnableMixin
from brq.log import logger


class Daemon:
    """
    Tool for daemonizing a consumer.
    """

    def __init__(self, *runnables: list[RunnableMixin]):
        self.runnables = runnables

    async def run_forever(self, stop_signals: list = [signal.SIGINT, signal.SIGTERM]):
        loop = asyncio.get_event_loop()

        stop_event = asyncio.Event()

        async def _stop():
            logger.debug("Signal received")
            stop_event.set()

        for sig in stop_signals:
            loop.add_signal_handler(sig, lambda: asyncio.create_task(_stop()))

        logger.info(f"Starting {len(self.runnables)} consumer...")
        for runnable in self.runnables:
            await runnable.start()
        logger.info(f"Consumer started, waiting for signals {stop_signals}...")
        await stop_event.wait()

        logger.info(f"Terminating consumer...")
        for runnable in self.runnables:
            await runnable.stop()
        logger.info(f"Consumer terminated")


if __name__ == "__main__":

    class DummyConsumer(RunnableMixin):
        async def initialize(self):
            print("init")

        async def run(self):
            print("run")
            await asyncio.sleep(1)

        async def cleanup(self):
            print("cleanup")

    asyncio.run(Daemon(DummyConsumer()).run_forever())



import redis.asyncio as redis

from brq.tools import (
    get_current_timestamp,
    get_current_timestamp_ms,
    timestamp_to_datetime,
)

STREAM_KEY = "stream"
DEAD_QUEUE_KEY = "dead"
DEFERRED_QUEUE_KEY = "deferred"


class RedisOperator:
    def __init__(
        self,
        redis_prefix: str = "brq",
        redis_seperator: str = ":",
    ):
        self.redis_prefix = redis_prefix
        self.redis_seperator = redis_seperator

    def get_redis_key(self, *key: list[str]) -> str:
        return self.redis_seperator.join([self.redis_prefix] + list(key))

    def get_stream_name(self, function_name: str) -> str:
        return self.get_redis_key("{%s}" % function_name, STREAM_KEY)

    def get_dead_message_key(self, function_name: str) -> str:
        return self.get_redis_key("{%s}" % function_name, DEAD_QUEUE_KEY)

    def get_deferred_key(self, function_name: str) -> str:
        return self.get_redis_key("{%s}" % function_name, DEFERRED_QUEUE_KEY)

    async def get_current_timestamp(self, redis_client: redis.Redis | redis.RedisCluster) -> int:
        return await get_current_timestamp(redis_client)

    async def get_current_timestamp_ms(self, redis_client: redis.Redis | redis.RedisCluster) -> int:
        return await get_current_timestamp_ms(redis_client)

    async def get_current_datetime(
        self, redis_client: redis.Redis | redis.RedisCluster, tz=None
    ) -> str:
        return timestamp_to_datetime(await get_current_timestamp(redis_client), tz=tz)



import asyncio

import pytest

from brq.browser import Browser
from brq.consumer import Consumer
from brq.producer import Producer


async def mock_consume(echo_str: str):
    print(echo_str)


async def delay_job(echo_str: str, delay_seconds: int = 1):
    await asyncio.sleep(delay_seconds)
    print(echo_str)


called = False


async def mock_consume_raise_exception(echo_str: str):
    global called
    if not called:
        called = True
        raise Exception("raise exception")
    else:
        print(echo_str)
        called = False


@pytest.mark.parametrize("run_parallel", [False, True])
async def test_consume_function(async_redis_client, capfd, run_parallel):
    producer = Producer(async_redis_client)
    consumer = Consumer(async_redis_client, mock_consume, run_parallel=run_parallel)
    browser = Browser(async_redis_client)
    await browser.status()
    await producer.run_job("mock_consume", ["hello"])
    await browser.status()
    await consumer.initialize()
    await browser.status()
    await consumer.run()
    await browser.status()

    out, err = capfd.readouterr()
    assert "hello" in out
    await consumer.cleanup()


async def test_count_unacked_jobs(async_redis_client):
    producer = Producer(async_redis_client)
    consumer = Consumer(async_redis_client, delay_job)
    browser = Browser(async_redis_client)
    assert await consumer.count_unacked_jobs("delay_job") == 0

    await browser.status()
    await producer.run_job("delay_job", ["hello"])
    await browser.status()
    await consumer.initialize()
    await browser.status()
    loop = asyncio.get_event_loop()
    loop.create_task(consumer.run())
    await asyncio.sleep(0.1)
    assert await consumer.count_unacked_jobs("delay_job") == 1
    await browser.status()
    await asyncio.sleep(1.1)
    assert await consumer.count_unacked_jobs("delay_job") == 0
    await consumer.cleanup()


async def test_consume_func_with_name(async_redis_client, capfd):
    producer = Producer(async_redis_client)
    consumer = Consumer(
        async_redis_client,
        mock_consume,
        register_function_name="mock_consume_with_name",
    )
    browser = Browser(async_redis_client)
    await browser.status()
    await producer.run_job("mock_consume_with_name", ["hello"])
    await browser.status()
    await consumer.initialize()
    await browser.status()
    await consumer.run()
    await browser.status()

    out, err = capfd.readouterr()
    assert "hello" in out
    await consumer.cleanup()


async def test_prune(async_redis_client, capfd):
    producer = Producer(async_redis_client)
    consumer = Consumer(async_redis_client, mock_consume)
    browser = Browser(async_redis_client)

    await browser.status()
    await producer.run_job("mock_consume", ["hello"])
    await browser.status()
    await producer.prune("mock_consume")
    await browser.status()
    await consumer.initialize()
    await browser.status()
    await consumer.run()
    await browser.status()
    out, err = capfd.readouterr()
    assert "hello" not in out
    await consumer.cleanup()


async def test_unprocessed_job(async_redis_client, capfd):
    producer = Producer(async_redis_client)
    consumer = Consumer(
        async_redis_client,
        mock_consume_raise_exception,
        process_timeout=0.001,
        retry_cooldown_time=1,
    )
    browser = Browser(async_redis_client)

    await browser.status()
    await producer.run_job("mock_consume_raise_exception", ["hello"])
    await browser.status()
    await consumer.initialize()
    await browser.status()
    await consumer.run()
    await browser.status()

    out, err = capfd.readouterr()
    assert "hello" not in out
    await asyncio.sleep(1.1)
    await consumer.run()
    await browser.status()

    out, err = capfd.readouterr()
    assert "hello" in out
    await consumer.cleanup()


@pytest.mark.parametrize("enable_dead_queue", [True, False])
async def test_process_dead_jobs(async_redis_client, capfd, enable_dead_queue):
    producer = Producer(async_redis_client)
    consumer = Consumer(
        async_redis_client,
        mock_consume_raise_exception,
        expire_time=0.001,
        enable_dead_queue=enable_dead_queue,
    )
    browser = Browser(async_redis_client)
    assert await producer.count_dead_messages("mock_consume_raise_exception") == 0

    await browser.status()
    await producer.run_job("mock_consume_raise_exception", ["hello"])
    await browser.status()
    await consumer.initialize()
    await browser.status()
    await consumer.run()
    await browser.status()
    out, err = capfd.readouterr()
    assert "hello" not in out
    await asyncio.sleep(0.1)
    await consumer._move_expired_jobs()
    if enable_dead_queue:
        assert await producer.count_dead_messages("mock_consume_raise_exception") == 1
    else:
        assert await producer.count_dead_messages("mock_consume_raise_exception") == 0
    await consumer.process_dead_jobs()
    await browser.status()

    out, err = capfd.readouterr()
    if enable_dead_queue:
        assert "hello" in out
    await consumer.cleanup()


async def test_deferred_job(async_redis_client, capfd):
    producer = Producer(async_redis_client)
    consumer = Consumer(async_redis_client, mock_consume)
    browser = Browser(async_redis_client)

    await browser.status()
    await producer.run_job("mock_consume", ["hello"], defer_seconds=2)
    await browser.status()
    assert await producer.get_deferred_jobs("mock_consume")
    await browser.status()
    await consumer.initialize()
    await browser.status()
    await consumer.run()
    await browser.status()
    out, err = capfd.readouterr()
    assert "hello" not in out
    await asyncio.sleep(2.1)
    await browser.status()
    await consumer.run()
    out, err = capfd.readouterr()
    assert "hello" in out
    await consumer.cleanup()


async def test_not_unique_deferred_job(
    async_redis_client,
    capfd,
):
    producer = Producer(async_redis_client)
    consumer = Consumer(async_redis_client, mock_consume)
    browser = Browser(async_redis_client)

    await browser.status()
    assert not await producer.deferred_job_exists("mock_consume", ["hello"])
    await producer.run_job("mock_consume", ["hello"], defer_seconds=2, unique=False)
    await producer.run_job("mock_consume", ["hello"], defer_seconds=2, unique=False)
    await browser.status()
    assert await producer.deferred_job_exists("mock_consume", ["hello"])
    assert len(await producer.get_deferred_jobs("mock_consume")) == 1
    await browser.status()
    await consumer.initialize()
    await browser.status()
    await consumer.run()
    await browser.status()
    out, err = capfd.readouterr()
    assert "hello" not in out

    await asyncio.sleep(2.1)
    await browser.status()
    await consumer.run()
    out, err = capfd.readouterr()
    assert "hello" in out
    await consumer.cleanup()


if __name__ == "__main__":
    pytest.main(["-s", "-vv", __file__])



from __future__ import annotations

from datetime import datetime
from typing import AsyncIterator, Optional, Union

import redis.asyncio as redis
from pydantic import BaseModel

from brq.log import logger
from brq.models import Job
from brq.operator import BrqOperator


def str_if_bytes(value: Union[str, bytes]) -> str:
    return value.decode("utf-8", errors="replace") if isinstance(value, bytes) else value


def pairs_to_dict(response, decode_keys=False, decode_string_values=False):
    """Create a dict given a list of key/value pairs"""
    if response is None:
        return {}
    if decode_keys or decode_string_values:
        # the iter form is faster, but I don't know how to make that work
        # with a str_if_bytes() map
        keys = response[::2]
        if decode_keys:
            keys = map(str_if_bytes, keys)
        values = response[1::2]
        if decode_string_values:
            values = map(str_if_bytes, values)
        return dict(zip(keys, values))
    else:
        it = iter(response)
        return dict(zip(it, it))


def prettify(function_name: str, job_queue_info: JobQueueInfo) -> str:
    stream_brief = (
        f"""
    length: {job_queue_info.stream_full_info.length}
    groups: {job_queue_info.stream_full_info.groups}
    radix_tree_keys: {job_queue_info.stream_full_info.radix_tree_keys}
    radix_tree_nodes: {job_queue_info.stream_full_info.radix_tree_nodes}
    last_generated_id: {job_queue_info.stream_full_info.last_generated_id}
    max_deleted_entry_id: {job_queue_info.stream_full_info.max_deleted_entry_id}
    entries_added: {job_queue_info.stream_full_info.entries_added}
    entries: {job_queue_info.stream_full_info.entries}
    group: {job_queue_info.stream_full_info.group}
""".strip()
        if job_queue_info.stream_full_info
        else "None"
    )

    template = f"""------------------------------------------------------
Function: {function_name}

  Brief for stream:
    {stream_brief}

  Brief for defer queue:
    length: {len(job_queue_info.defer_queue_info)}
    details: {job_queue_info.defer_queue_info}

  Brief for dead queue:
    length: {len(job_queue_info.dead_queue_info)}
    details: {job_queue_info.dead_queue_info}
"""

    return template.strip() + "\n"


class ScannerMixin(BrqOperator):
    def __init__(
        self,
        redis: redis.Redis | redis.RedisCluster,
        redis_prefix: str = "brq",
        redis_seperator: str = ":",
    ):
        self.redis = redis
        self.redis_prefix = redis_prefix
        self.redis_seperator = redis_seperator

    def scan_iter(self) -> AsyncIterator[str]:
        if self.redis_prefix == "*":
            scan_key = "*"
        else:
            scan_key = self.get_redis_key("*")
        logger.debug(f"Scanning {scan_key}")
        return self.redis.scan_iter(match=scan_key)

    async def parse_redis_key(self, redis_key: str) -> tuple[str, str]:
        function_and_domain = redis_key[len(self.redis_prefix + self.redis_seperator) :]
        function, domain_key = function_and_domain.split(self.redis_seperator)

        if "{" in function:
            function = function.replace("{", "").replace("}", "")

        return function, domain_key

    async def get_stream_full_info(self, stream_key: str) -> _StreamFullInfo | None:
        if not await self.redis.exists(stream_key):
            return None

        try:
            full_response = await self.redis.xinfo_stream(stream_key, full=True)
        except IndexError:
            # https://github.com/redis/redis-py/pull/3282
            logger.warning(
                "Due to redis-py bug, only brief info when no consumer group found. Please upgrade redis>=5.0.7 to fix this."
            )
            full_response = await self.redis.xinfo_stream(stream_key)

        return _StreamFullInfo.from_xinfo_stream_full(full_response)


class _StreamFullInfo(BaseModel):
    length: int
    groups: int
    radix_tree_keys: int
    radix_tree_nodes: int
    last_generated_id: str
    max_deleted_entry_id: Optional[str] = None
    entries_added: Optional[int] = None
    entries: dict[str, Job]
    group: dict[str, _GroupInfo]

    @classmethod
    def from_xinfo_stream_full(cls, full_response: dict) -> _StreamFullInfo:
        entries = (
            {k: Job.model_validate_json(v["payload"]) for k, v in full_response["entries"].items()}
            if "entries" in full_response
            else {}
        )
        return cls(
            length=full_response["length"],
            groups=(
                len(full_response["groups"])
                if isinstance(full_response["groups"], list)
                else full_response["groups"]
            ),
            radix_tree_keys=full_response["radix-tree-keys"],
            radix_tree_nodes=full_response["radix-tree-nodes"],
            last_generated_id=full_response["last-generated-id"],
            max_deleted_entry_id=full_response.get("max-deleted-entry-id"),
            entries_added=full_response.get("entries-added"),
            entries=entries,
            group=(
                {g["name"]: _GroupInfo.from_xinfo_stream_full(g) for g in full_response["groups"]}
                if isinstance(full_response["groups"], list)
                else {}
            ),
        )


class _GroupInfo(BaseModel):
    name: str
    last_delivered_id: str
    entries_read: Optional[int] = None
    lag: Optional[int] = None
    pel_count: int
    pending: list[_PendingJobInfo]
    consumers: list[_ConsumerInfo]

    @classmethod
    def from_xinfo_stream_full(cls, full_response: dict) -> _GroupInfo:
        return cls(
            name=full_response["name"],
            last_delivered_id=full_response["last-delivered-id"],
            entries_read=full_response.get("entries-read"),
            lag=full_response.get("lag"),
            pel_count=full_response["pel-count"],
            pending=[_PendingJobInfo.from_xinfo_stream_full(p) for p in full_response["pending"]],
            consumers=[_ConsumerInfo.from_xinfo_stream_full(c) for c in full_response["consumers"]],
        )


class _PendingJobInfo(BaseModel):
    id: str
    consumer: str
    delivered_at: datetime
    delivered_times: int

    @classmethod
    def from_xinfo_stream_full(cls, full_response: dict) -> _PendingJobInfo:
        return cls(
            id=full_response[0],
            consumer=full_response[1],
            delivered_at=datetime.fromtimestamp(full_response[2] / 1000),
            delivered_times=full_response[3],
        )


class _ConsumerInfo(BaseModel):
    """
    Note that before Redis 7.2.0, seen-time used to denote the last successful interaction.
    In 7.2.0, active-time was added and seen-time was changed to denote the last attempted interaction.
    """

    name: str
    seen_time: datetime
    active_time: Optional[datetime] = None
    pel_count: int
    pending: list[_PendingJobInfo]

    @classmethod
    def from_xinfo_stream_full(cls, full_response: dict | list) -> _ConsumerInfo:
        if isinstance(full_response, list):
            # Incase redis-py not parse consumer
            full_response = pairs_to_dict(full_response, decode_keys=True)

        active_time = full_response.get("active-time")
        if active_time:
            active_time = datetime.fromtimestamp(active_time / 1000)
        else:
            active_time = None
        return cls(
            name=full_response["name"],
            seen_time=datetime.fromtimestamp(full_response["seen-time"] / 1000),
            active_time=active_time,
            pel_count=full_response["pel-count"],
            pending=[
                _PendingJobInfo(
                    id=p[0],
                    consumer=full_response["name"],
                    delivered_at=p[1],
                    delivered_times=p[2],
                )
                for p in full_response["pending"]
            ],
        )


class DeferQueueInfo(BaseModel):
    defer_until: datetime
    job: Job


class DeadQueueInfo(BaseModel):
    dead_at: datetime
    job: Job


class JobQueueInfo(BaseModel):
    stream_full_info: Optional[_StreamFullInfo]
    defer_queue_info: list[DeferQueueInfo]
    dead_queue_info: list[DeadQueueInfo]


class BrqJobsInfo(BaseModel):
    job_queue_info: dict[str, JobQueueInfo]


class Browser(ScannerMixin):
    def __init__(
        self,
        redis: redis.Redis | redis.RedisCluster,
        redis_prefix: str = "brq",
        redis_seperator: str = ":",
    ):
        super().__init__(redis, redis_prefix, redis_seperator)

    async def status(self) -> BrqJobsInfo:
        inspect_functions = []
        async for stream_key in self.scan_iter():
            function_name, _ = await self.parse_redis_key(stream_key)
            (
                inspect_functions.append(function_name)
                if function_name not in inspect_functions
                else None
            )
        return BrqJobsInfo(
            job_queue_info={
                function_name: await self.one_queue_info(function_name)
                for function_name in inspect_functions
            }
        )

    async def one_queue_info(self, function_name: str) -> JobQueueInfo:
        stream_key = self.get_stream_name(function_name)
        dead_job_info = [
            DeadQueueInfo(
                dead_at=k,
                job=v,
            )
            for k, v in (await self.get_dead_messages(function_name)).items()
        ]

        defer_job_info = [
            DeferQueueInfo(
                defer_until=k,
                job=v,
            )
            for k, v in (await self.get_deferred_jobs(function_name)).items()
        ]

        return JobQueueInfo(
            stream_full_info=await self.get_stream_full_info(stream_key),
            defer_queue_info=defer_job_info,
            dead_queue_info=dead_job_info,
        )



from datetime import datetime
from typing import Any

import redis.asyncio as redis

from brq.log import logger
from brq.models import Job
from brq.rds import RedisOperator


class BrqOperator(RedisOperator):
    def __init__(
        self,
        redis: redis.Redis | redis.RedisCluster,
        redis_prefix: str = "brq",
        redis_seperator: str = ":",
    ):
        super().__init__(redis_prefix, redis_seperator)
        self.redis = redis

    async def enque_deferred_job(self, function_name: str, maxlen: int = 1000):
        """
        From zset to stream
        """
        lua_script = """
        local zset_key = KEYS[1]
        local stream_key = KEYS[2]
        local current_timestamp = ARGV[1]
        local maxlen = ARGV[2]

        local elements = redis.call('ZRANGEBYSCORE', zset_key, '-inf', current_timestamp)

        for i, element in ipairs(elements) do
            redis.call('ZREM', zset_key, element)
            redis.call('XADD', stream_key, 'MAXLEN', maxlen, '*', 'payload', element)
        end

        return elements
        """
        defer_key = self.get_deferred_key(function_name)
        stream_name = self.get_stream_name(function_name)

        elements = await self.redis.eval(
            lua_script,
            2,
            defer_key,
            stream_name,
            await self.get_current_timestamp_ms(self.redis),
            maxlen,
        )
        if elements:
            logger.debug(f"Enqueued deferred jobs: {elements}")

    async def _remove_deferred_job(
        self,
        function_name: str,
        job: Job,
    ):
        """
        Remove specific deferred job

        Args:
            function_name (str): function name
            job (Job): job to be removed
        """
        defer_key = self.get_deferred_key(function_name)
        delete_nums = await self.redis.zrem(defer_key, job.to_redis())
        logger.info(f"Removed {delete_nums} deferred jobs")

    async def get_defer_timestamp_ms(
        self,
        defer_until: datetime = None,
        defer_hours: int = 0,
        defer_minutes: int = 0,
        defer_seconds: int = 0,
    ):
        if not any(
            [
                defer_until,
                defer_hours,
                defer_minutes,
                defer_seconds,
            ]
        ):
            return None

        if defer_until:
            logger.debug(f"Using defer_until, ignore defer_hours, defer_minutes, defer_seconds")
            defer_until = defer_until.timestamp()

        else:
            defer_until = (
                await self.get_current_timestamp(self.redis)
                + defer_hours * 60 * 60
                + defer_minutes * 60
                + defer_seconds
            )

        return defer_until * 1000

    async def get_deferred_jobs(self, function_name: str) -> dict[datetime:Job]:
        """
        Get all deferred jobs

        Args:
            function_name (str): function name

        Returns:
            dict[datetime:Job]: deferred jobs,
                key is the datetime when the job expected to be executed
                value is the job
        """
        defer_key = self.get_deferred_key(function_name)
        return {
            datetime.fromtimestamp(float(element[1]) / 1000): Job.from_redis(element[0])
            for element in await self.redis.zrangebyscore(
                defer_key, "-inf", "+inf", withscores=True
            )
        }

    async def get_dead_messages(self, function_name: str) -> dict[datetime:Job]:
        """
        Get all dead messages

        Args:
            function_name (str): function name

        Returns:
            dict[datetime:Job]: dead messages,
                key is the datetime when the job to be moved to dead message
                value is the job
        """
        dead_key = self.get_dead_message_key(function_name)
        return {
            datetime.fromtimestamp(float(element[1]) / 1000): Job.from_redis(element[0])
            for element in await self.redis.zrangebyscore(dead_key, "-inf", "+inf", withscores=True)
        }

    async def count_deferred_jobs(self, function_name: str) -> int:
        defer_key = self.get_deferred_key(function_name)
        if not await self.redis.exists(defer_key):
            return 0
        return await self.redis.zcard(defer_key)

    async def count_stream(self, function_name: str) -> int:
        """
        Count stream length

        If Consumer's `delete_message_after_process` if False, will include already processed messages
        """
        stream_name = self.get_stream_name(function_name)
        if not await self.redis.exists(stream_name):
            return 0
        return await self.redis.xlen(stream_name)

    async def count_unacked_jobs(
        self, function_name: str, group_name: str = "default-workers"
    ) -> int:
        """
        Count unacked jobs in group

        Args:
            function_name (str): function name
            group_name (str, optional): group name. Defaults to "default-workers". Should be the same as Consumer's `group_name`
        """
        stream_name = self.get_stream_name(function_name)
        if not await self.redis.exists(stream_name):
            return 0
        try:
            consumer_group_info = await self.redis.xinfo_consumers(stream_name, group_name)
            return consumer_group_info[0]["pending"]
        except redis.ResponseError as e:
            if "NOGROUP No such consumer group" in e.args[0]:
                return 0
            raise

    async def count_dead_messages(self, function_name: str) -> int:
        dead_key = self.get_dead_message_key(function_name)
        if not await self.redis.exists(dead_key):
            return 0
        return await self.redis.zcard(dead_key)

    async def emit_deferred_job(self, function_name: str, defer_until: int, job: Job):
        defer_key = self.get_deferred_key(function_name)
        await self.redis.zadd(defer_key, {job.to_redis(): defer_until})
        return job

    async def _deferred_job_exists(
        self,
        function_name: str,
        job: Job,
    ) -> bool:
        defer_key = self.get_deferred_key(function_name)
        return await self.redis.zscore(defer_key, job.to_redis())



import asyncio
import copy
import os
import uuid
from typing import Any, Awaitable, Callable

import redis.asyncio as redis

from brq.log import logger
from brq.models import Job
from brq.operator import BrqOperator
from brq.tools import event_wait

CONSUMER_IDENTIFIER_ENV = "BRQ_CONSUMER_IDENTIFIER"


class RunnableMixin:
    def __init__(self):
        self._stop_event = asyncio.Event()
        self._task: None | asyncio.Task = None

    async def start(self):
        if self._task:
            raise RuntimeError("Consumer already started.")
        self._stop_event.clear()
        self._task = asyncio.create_task(self._start())

    async def _start(self):
        try:
            await self.initialize()
        except Exception as e:
            # Unrecoverable error
            logger.exception(e)
            exit(1)
        while not await event_wait(self._stop_event, 0.1):
            if self._stop_event.is_set():
                break
            try:
                await self.run()
            except Exception as e:
                logger.exception(e)

        await self.cleanup()

    async def stop(self):
        self._stop_event.set()
        if self._task:
            logger.info("Waiting for consumer finish...")
            await self._task

    async def initialize(self):
        """
        Initialize, implement in subclass
        """

    async def run(self):
        """
        Run, implement in subclass
        """

    async def cleanup(self):
        """
        Cleanup, implement in subclass
        """


class Consumer(BrqOperator, RunnableMixin):
    """
    A consumer for redis stream.

    When many replicas of one kind of consumer are created, they should be in **the same** group.
    When many kind of consumers are created, they should be in **different** groups.

    See [example](https://github.com/Wh1isper/brq/tree/main/examples) for running example.

    Args:
        redis(redis.Redis | redis.RedisCluster): Redis client.
        awaitable_function(Callable[[Any], Awaitable[Any]]): The function to be called.
        register_function_name(str): The function name to be registered, if None, will use the function name of `awaitable_function`.
        group_name(str, default="default-workers"): The group name. All replicas should be in the same group.
        consumer_identifier(str, default=os.getenv(CONSUMER_IDENTIFIER_ENV, uuid.uuid4().hex)): The consumer identifier. For restarting in some special case.
        count_per_fetch(int, default=1): The count of message per fetch.
        block_time(int, default=1): The block time of reading stream.
        expire_time(int, default=60*60): The expire time of a message. Expired messages will be moved to dead queue.
        process_timeout(int, default=60): The timeout of a job. If a job is not finished in this time, it will be reprocessed.
        retry_lock_time(int, default=300): The lock time for retrying. Only one consumer can retry jobs at a time.
        retry_cooldown_time(int, default=60): The cooldown time between retrying.
        redis_prefix(str, default="brq"): The prefix of redis key.
        redis_seperator(str, default=":"): The seperator of redis key.
        enable_enque_deferred_job(bool, default=True): Whether to enable enque deferred job. If not, this consumer won't enque deferred jobs.
        enable_reprocess_timeout_job(bool, default=True): Whether to enable reprocess timeout job. If not, this consumer won't reprocess timeout jobs, jobs with expired timeout will be moved to dead queue(if enabled).
        enable_dead_queue(bool, default=True): Whether to enable dead queue. If not, this consumer won't move expired jobs to dead queue but just delete them.
        max_message_len(int, default=1000): The maximum length of a message. Follow redis stream `maxlen`.
        delete_message_after_process(bool, default=False): Whether to delete message after process. If many consumer groups are used, this should be set to False.
        run_parallel(bool, default=False): Whether to run in parallel.
    """

    def __init__(
        self,
        redis: redis.Redis | redis.RedisCluster,
        awaitable_function: Callable[[Any], Awaitable[Any]],
        register_function_name: str = None,
        group_name="default-workers",
        consumer_identifier: str = os.getenv(CONSUMER_IDENTIFIER_ENV, uuid.uuid4().hex),
        count_per_fetch: int = 1,
        block_time: int = 1,
        expire_time: int = 60 * 60,
        process_timeout: int = 60,
        retry_lock_time: int = 300,
        retry_cooldown_time: int = 60,
        redis_prefix: str = "brq",
        redis_seperator: str = ":",
        enable_enque_deferred_job: bool = True,
        enable_reprocess_timeout_job: bool = True,
        enable_dead_queue: bool = True,
        max_message_len: int = 1000,
        delete_message_after_process: bool = False,
        run_parallel: bool = False,
    ):
        super().__init__(redis, redis_prefix, redis_seperator)
        self._stop_event = asyncio.Event()
        self._task: None | asyncio.Task = None

        self.awaitable_function = awaitable_function
        self.register_function_name = register_function_name or awaitable_function.__name__
        self.group_name = group_name
        self.consumer_identifier = consumer_identifier
        self.count_per_fetch = count_per_fetch

        self.block_time = block_time
        self.expire_time = expire_time
        self.process_timeout = process_timeout

        if self.expire_time < self.process_timeout:
            logger.warning(
                f"expire_time({self.expire_time}) < process_timeout({self.process_timeout}), will causes no job retried but moved dead queue(if enabled)."
            )

        self.retry_lock_time = retry_lock_time
        self.retry_cooldown_time = retry_cooldown_time

        self.enable_enque_deferred_job = enable_enque_deferred_job
        self.enable_reprocess_timeout_job = enable_reprocess_timeout_job
        self.enable_dead_queue = enable_dead_queue
        self.max_message_len = max_message_len
        self.delete_message_after_process = delete_message_after_process
        self.run_parallel = run_parallel

    @property
    def retry_lock_key(self) -> str:
        return f"{self.stream_name}-{self.group_name}-lock"

    @property
    def retry_cooldown_key(self) -> str:
        return f"{self.stream_name}-{self.group_name}-cooldown"

    @property
    def stream_name(self) -> str:
        return self.get_stream_name(self.register_function_name)

    @property
    def deferred_key(self) -> str:
        return self.get_deferred_key(self.register_function_name)

    @property
    def dead_key(self) -> str:
        return self.get_dead_message_key(self.register_function_name)

    async def _is_retry_cooldown(self) -> bool:
        return await self.redis.exists(self.retry_cooldown_key)

    async def _set_retry_timeout(self):
        await self.redis.set(self.retry_cooldown_key, "1", ex=self.retry_cooldown_time)

    async def _consume(self):
        try:
            if not await self._is_retry_cooldown() and await self._acquire_retry_lock():
                await self._process_unacked_job()
                await self._move_expired_jobs()
                await self._set_retry_timeout()
        finally:
            await self._release_retry_lock()

        if self.run_parallel:
            await self._pool_job_prallel()
        else:
            await self._pool_job()

    async def process_dead_jobs(self, raise_exception: bool = False):
        for job in (Job.from_redis(j) for j in await self.redis.zrange(self.dead_key, 0, -1)):
            try:
                await self.awaitable_function(*job.args, **job.kwargs)
            except Exception as e:
                logger.exception(e)
                if raise_exception:
                    raise e
            else:
                await self.redis.zrem(self.dead_key, job.to_redis())

    async def _move_expired_jobs(self):
        while True:
            if self._stop_event.is_set():
                break

            message_id = "0-0"
            expired_messages = (
                await self.redis.xautoclaim(
                    self.stream_name,
                    self.group_name,
                    self.consumer_identifier,
                    min_idle_time=int(self.expire_time * 1000),
                    start_id=message_id,
                    count=100,
                )
            )[1]
            if not expired_messages:
                break

            for message_id, serialized_job in expired_messages:
                if message_id == serialized_job == None:
                    # Fix (None, None) for redis 6.x
                    continue
                job = Job.from_message(serialized_job)
                if self.enable_dead_queue:
                    await self.redis.zadd(self.dead_key, {job.to_redis(): job.create_at})
                    logger.info(f"Put expired job {job} to dead queue")
                await self.redis.xdel(self.stream_name, message_id)
                logger.debug(f"{job} expired")

    async def _process_unacked_job(self):
        if not self.enable_reprocess_timeout_job:
            return

        while True:
            if self._stop_event.is_set():
                break

            message_id = "0-0"
            claimed_messages = (
                await self.redis.xautoclaim(
                    self.stream_name,
                    self.group_name,
                    self.consumer_identifier,
                    min_idle_time=int(self.process_timeout * 1000),
                    start_id=message_id,
                    count=10,
                )
            )[1]
            if not claimed_messages:
                break

            for message_id, serialized_job in claimed_messages:
                if message_id == serialized_job == None:
                    # Fix (None, None) for redis 6.x
                    continue
                job = Job.from_message(serialized_job)
                try:
                    await self.awaitable_function(*job.args, **job.kwargs)
                except Exception as e:
                    logger.exception(e)
                else:
                    logger.info(f"Retry {job} successfully")
                    await self.redis.xack(self.stream_name, self.group_name, message_id)

                    if self.delete_message_after_process:
                        await self.redis.xdel(self.stream_name, message_id)

    async def _pool_job(self):
        poll_result = await self.redis.xreadgroup(
            groupname=self.group_name,
            consumername=self.consumer_identifier,
            streams={self.stream_name: ">"},
            count=self.count_per_fetch,
            block=self.block_time,
        )
        if not poll_result:
            return

        _, messages = poll_result[0]
        for message_id, serialized_job in messages:
            job = Job.from_message(serialized_job)
            try:
                await self.awaitable_function(*job.args, **job.kwargs)
            except Exception as e:
                logger.exception(e)
            else:
                await self.redis.xack(self.stream_name, self.group_name, message_id)

                if self.delete_message_after_process:
                    await self.redis.xdel(self.stream_name, message_id)

    async def _pool_job_prallel(self):
        pool_result = await self.redis.xreadgroup(
            groupname=self.group_name,
            consumername=self.consumer_identifier,
            streams={self.stream_name: ">"},
            count=self.count_per_fetch,
            block=self.block_time,
        )
        if not pool_result:
            return
        _, messages = pool_result[0]

        async def _job_wrap(message_id, *args, **kwargs):
            await self.awaitable_function(*args, **kwargs)
            return message_id

        jobs = {
            messages_id: Job.from_message(serialized_job)
            for messages_id, serialized_job in messages
        }
        results = await asyncio.gather(
            *[_job_wrap(message_id, *job.args, **job.kwargs) for message_id, job in jobs.items()],
            return_exceptions=True,
        )

        for message_id, result in zip(jobs, results):
            if isinstance(result, Exception):
                logger.exception(result)
                continue

            await self.redis.xack(self.stream_name, self.group_name, message_id)
            if self.delete_message_after_process:
                await self.redis.xdel(self.stream_name, message_id)

    async def _acquire_retry_lock(self) -> bool:
        return await self.redis.get(
            self.retry_lock_key
        ) == self.consumer_identifier or await self.redis.set(
            self.retry_lock_key,
            self.consumer_identifier,
            ex=self.retry_lock_time,
            nx=True,
        )

    async def _release_retry_lock(self) -> bool:
        lua_script = """
            if redis.call("get", KEYS[1]) == ARGV[1] then
                return redis.call("del", KEYS[1])
            else
                return 0
            end
        """

        deleted = bool(
            await self.redis.eval(lua_script, 1, self.retry_lock_key, self.consumer_identifier)
        )
        return deleted

    async def initialize(self):
        try:
            logger.info(
                f"Creating consumer group: {self.group_name} for stream: {self.stream_name}"
            )
            await self.redis.xgroup_create(self.stream_name, self.group_name, 0, mkstream=True)
        except redis.ResponseError as e:
            if e.args[0] != "BUSYGROUP Consumer Group name already exists":
                raise
            logger.info(f"Consumer group already exists: {self.group_name}, skipping...")

    async def run(self):
        if self.enable_enque_deferred_job:
            await self.enque_deferred_job(self.register_function_name, self.max_message_len)
        await self._consume()

    async def cleanup(self):
        await self.redis.xgroup_delconsumer(
            self.stream_name, self.group_name, self.consumer_identifier
        )
        if await self._release_retry_lock():
            logger.info(f"Stop reprocessing unprocessed events: {self.consumer_identifier}")



REDIS_URL_ENV = "REDIS_URL"
REDIS_HOST_ENV = "REDIS_HOST"
REDIS_PORT_ENV = "REDIS_PORT"
REDIS_DB_ENV = "REDIS_DB"
REDIS_CLUSTER_ENV = "REDIS_CLUSTER"
REDIS_TLS_ENV = "REDIS_TLS"
REDIS_USERNAME_ENV = "REDIS_USERNAME"
REDIS_PASSWORD_ENV = "REDIS_PASSWORD"


REDIS_PREFIX_ENV = "REDIS_PREFIX"
REDIS_SEPERATOR_ENV = "REDIS_SEPERATOR"



import os

from brq.consumer import Consumer
from brq.daemon import Daemon
from brq.tools import get_redis_client, get_redis_url


async def echo(message):
    print(message)


async def main():
    redis_url = get_redis_url(
        host=os.getenv("REDIS_HOST", "localhost"),
        port=int(os.getenv("REDIS_PORT", 6379)),
        db=int(os.getenv("REDIS_DB", 0)),
        cluster=bool(os.getenv("REDIS_CLUSTER", False)),
        tls=bool(os.getenv("REDIS_TLS", False)),
    )
    async with get_redis_client(redis_url) as async_redis_client:
        await Consumer(async_redis_client, echo).process_dead_jobs()


if __name__ == "__main__":
    import asyncio

    asyncio.run(main())



import asyncio
from functools import wraps

import click

from brq.browser import Browser, prettify
from brq.tools import get_redis_client, get_redis_url

from .envs import *


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@click.command()
@click.option(
    "--redis-host",
    default="localhost",
    help="Redis host",
    type=str,
    envvar=REDIS_HOST_ENV,
)
@click.option(
    "--redis-port",
    default=6379,
    help="Redis port",
    type=int,
    envvar=REDIS_PORT_ENV,
)
@click.option(
    "--redis-db",
    default=0,
    help="Redis db",
    type=int,
    envvar=REDIS_DB_ENV,
)
@click.option(
    "--redis-cluster",
    default=False,
    help="Redis cluster",
    type=bool,
    envvar=REDIS_CLUSTER_ENV,
)
@click.option(
    "--redis-tls",
    default=False,
    help="Redis TLS",
    type=bool,
    envvar=REDIS_TLS_ENV,
)
@click.option(
    "--redis-username",
    default="",
    help="Redis username",
    type=str,
    envvar=REDIS_USERNAME_ENV,
)
@click.option(
    "--redis-password",
    default="",
    help="Redis password",
    type=str,
    envvar=REDIS_PASSWORD_ENV,
)
@click.option("--redis-url", required=False, help="Redis URL", type=str, envvar=REDIS_URL_ENV)
@click.option(
    "--redis-prefix",
    required=False,
    help="Redis prefix",
    type=str,
    envvar=REDIS_PREFIX_ENV,
)
@click.option(
    "--redis-seperator",
    required=False,
    help="Redis seperator",
    type=str,
    envvar=REDIS_SEPERATOR_ENV,
)
@coro
async def browser(
    redis_host,
    redis_port,
    redis_db,
    redis_cluster,
    redis_tls,
    redis_username,
    redis_password,
    redis_url,
    redis_prefix,
    redis_seperator,
):
    """
    See all job and status in CLI
    """

    redis_url = redis_url or get_redis_url(
        host=redis_host,
        port=redis_port,
        db=redis_db,
        cluster=redis_cluster,
        tls=redis_tls,
        username=redis_username,
        password=redis_password,
    )
    async with get_redis_client(redis_url, is_cluster=redis_cluster) as redis_client:
        brq_jobs = await Browser(
            redis_client,
            **{
                k: v
                for k, v in dict(
                    redis_prefix=redis_prefix,
                    redis_seperator=redis_seperator,
                ).items()
                if v is not None
            },
        ).status()

        print(
            f"{len(brq_jobs.job_queue_info)} functions in total: {list(brq_jobs.job_queue_info.keys())}"
        )
        for function_name, job_queue_info in brq_jobs.job_queue_info.items():
            print(prettify(function_name, job_queue_info))


@click.group()
def cli():
    pass


cli.add_command(browser)



from datetime import datetime
from typing import Any
from uuid import uuid4

import redis.asyncio as redis

from brq.log import logger
from brq.models import Job
from brq.operator import BrqOperator


class Producer(BrqOperator):
    """
    Producer to publish jobs to redis stream

    Use `max_message_len` to control max message length.

    Job control methods:
    * Call `run_job` to emit job.
    * Call `prune` to remove all jobs.

    Queue methods:
    * Call `count_stream` to count all jobs.
    * Call `count_deferred_jobs` to count deferred jobs.
    * Call `count_unacked_jobs` to count unacked jobs.
    * Call `count_dead_messages` to count all dead messages.

    Some higher level methods, see ``BrqOperator``

    For inspect the whole brq, see ``browser.py``

    Args:
        redis (redis.Redis | redis.RedisCluster): async redis client
        redis_prefix (str, optional): redis prefix. Defaults to "brq".
        redis_seperator (str, optional): redis seperator. Defaults to ":".
        max_message_len (int, optional): max message length. Defaults to 1000. Follow redis stream `maxlen`.
    """

    def __init__(
        self,
        redis: redis.Redis | redis.RedisCluster,
        redis_prefix: str = "brq",
        redis_seperator: str = ":",
        max_message_len: int = 1000,
    ):
        super().__init__(redis, redis_prefix, redis_seperator)
        self.max_message_len = max_message_len

    async def run_job(
        self,
        function_name: str,
        args: list[Any] = None,
        kwargs: dict[str, Any] = None,
        defer_until: datetime = None,
        defer_hours: int = 0,
        defer_minutes: int = 0,
        defer_seconds: int = 0,
        unique: bool = True,
    ) -> Job:
        """
        Emit job to redis stream. The args and kwargs will be serialized to json.

        If `defer_until` is not None, job will be deferred until `defer_until`.

        Else if any of `defer_hours`, `defer_minutes`, `defer_seconds` is not 0,
        job will be deferred for `defer_hours` hours, `defer_minutes` minutes, `defer_seconds` seconds.

        Args:
            function_name (str): function name
            args (list, optional): args. Defaults to None.
            kwargs (dict, optional): kwargs. Defaults to None.
            defer_until (datetime, optional): defer until. Defaults to None.
            defer_hours (int, optional): defer hours. Defaults to 0.
            defer_minutes (int, optional): defer minutes. Defaults to 0.
            defer_seconds (int, optional): defer seconds. Defaults to 0.
            unique (bool, optional): The deferred job will be unique. Defaults to True.

        Example:
            >>> await producer.run_job('function_name', args=[], kwargs={})

        Returns:
            Job: created job
        """

        args = args or []
        kwargs = kwargs or {}

        defer_until = await self.get_defer_timestamp_ms(
            defer_until, defer_hours, defer_minutes, defer_seconds
        )

        if defer_until:
            logger.debug(
                f"Deferring job: {function_name} until {datetime.fromtimestamp(defer_until / 1000)}"
            )
            job = await self.run_deferred_job(
                function_name,
                defer_until,
                args,
                kwargs,
                unique=unique,
            )
        else:
            logger.debug(f"Scheduling job: {function_name}")
            job = await self._run_job(function_name, args, kwargs)

        return job

    async def _run_job(
        self, function_name: str, args: list[Any] = None, kwargs: dict[str, Any] = None
    ) -> Job:
        stream_name = self.get_stream_name(function_name)
        created_at = await self.get_current_timestamp_ms(self.redis)

        job = Job(
            args=args or [],
            kwargs=kwargs or {},
            create_at=created_at,
        )
        await self.redis.xadd(stream_name, job.to_message(), maxlen=self.max_message_len)
        return job

    async def run_deferred_job(
        self,
        function_name: str,
        defer_until: int,
        args: list[Any] = None,
        kwargs: dict[str, Any] = None,
        unique: bool = True,
    ) -> Job:
        if unique:
            job = Job(
                args=args or [],
                kwargs=kwargs or {},
                create_at=await self.get_current_timestamp_ms(self.redis),
                uid=uuid4().hex,
            )
        else:
            job = Job(
                args=args or [],
                kwargs=kwargs or {},
            )
        return await self.emit_deferred_job(function_name, defer_until, job)

    async def prune(self, function_name: str):
        """
        Prune all jobs for function_name

        Args:
            function_name (str): function name
        """
        stream_name = self.get_stream_name(function_name)
        defer_key = self.get_deferred_key(function_name)
        dead_key = self.get_dead_message_key(function_name)
        await self.redis.delete(stream_name, defer_key, dead_key)

    async def remove_deferred_job(
        self,
        function_name: str,
        args: list[Any] = None,
        kwargs: dict[str, Any] = None,
    ):
        """
        Only not unique job can be removed
        """

        job = Job(
            args=args or [],
            kwargs=kwargs or {},
            create_at=0,
        )
        await self._remove_deferred_job(function_name, job)

    async def deferred_job_exists(
        self, function_name: str, args: list[Any] = None, kwargs: dict[str, Any] = None
    ) -> bool:
        """
        Only not unique job can be detected
        """
        job = Job(
            args=args or [],
            kwargs=kwargs or {},
            create_at=0,
        )
        return await self._deferred_job_exists(function_name, job)
